import 'package:flutter/material.dart';
import 'package:flutter_custom_toolbox/flutter_toolbox.dart';

import 'package:momomotus/models/game/game.dart';
import 'package:momomotus/models/settings/settings_game.dart';
import 'package:momomotus/models/settings/settings_global.dart';

part 'game_state.dart';

class GameCubit extends HydratedCubit<GameState> {
  GameCubit()
      : super(GameState(
          currentGame: Game.createEmpty(),
        ));

  void updateState(Game game) {
    emit(GameState(
      currentGame: game,
    ));
  }

  void refresh() {
    final Game game = Game(
      // Settings
      gameSettings: state.currentGame.gameSettings,
      globalSettings: state.currentGame.globalSettings,
      // State
      isRunning: state.currentGame.isRunning,
      isStarted: state.currentGame.isStarted,
      animationInProgress: state.currentGame.animationInProgress,
      // Base data
      word: state.currentGame.word,
      dictionary: state.currentGame.dictionary,
      // Game data
      currentGuess: state.currentGame.currentGuess,
      foundLetters: state.currentGame.foundLetters,
      guesses: state.currentGame.guesses,
      foundWord: state.currentGame.foundWord,
    );
    // game.dump();

    updateState(game);
  }

  void startNewGame({
    required GameSettings gameSettings,
    required GlobalSettings globalSettings,
  }) {
    final Game newGame = Game.createNew(
      gameSettings: gameSettings,
      globalSettings: globalSettings,
    );

    newGame.dump();

    updateState(newGame);
    refresh();
  }

  void quitGame() {
    state.currentGame.isRunning = false;
    refresh();
  }

  void resumeSavedGame() {
    state.currentGame.isRunning = true;
    refresh();
  }

  void deleteSavedGame() {
    state.currentGame.isRunning = false;
    refresh();
  }

  void currentGuessAddLetter(String letter) {
    if (state.currentGame.currentGuess.length <
        int.parse(state.currentGame.gameSettings.length)) {
      state.currentGame.currentGuess = state.currentGame.currentGuess + letter;
      refresh();
    }
  }

  void currentGuessRemoveLetter() {
    if (state.currentGame.currentGuess.isNotEmpty) {
      state.currentGame.currentGuess = state.currentGame.currentGuess
          .substring(0, state.currentGame.currentGuess.length - 1);
      refresh();
    }
  }

  void currentGuessSubmitWord() {
    state.currentGame.isStarted = true;

    if (state.currentGame.currentGuess.length ==
        int.parse(state.currentGame.gameSettings.length)) {
      if (state.currentGame.currentGuess == state.currentGame.word) {
        printlog('Word found!');
        state.currentGame.foundWord = true;
      }
      addGuess(state.currentGame.currentGuess);
      refresh();
    }
  }

  void currentGuessSubmitWrongWord() {
    addGuess(state.currentGame.currentGuess);
    refresh();
  }

  void addFoundLetter(String letter, int i) {
    state.currentGame.foundLetters = state.currentGame.foundLetters.substring(0, i) +
        letter +
        state.currentGame.foundLetters.substring(i + 1);
  }

  void addGuess(String word) {
    state.currentGame.guesses.add(word);
    state.currentGame.currentGuess = '';
  }

  void submitWord() {
    if (state.currentGame.checkWordIsValid(state.currentGame.currentGuess)) {
      printlog('Ok word allowed: "${state.currentGame.currentGuess}".');
      currentGuessSubmitWord();
    } else {
      printlog('Unknown word: "${state.currentGame.currentGuess}".');
      currentGuessSubmitWrongWord();
    }
  }

  List<String> getTips(String candidate) {
    String word = state.currentGame.word;
    final int wordLength = int.parse(state.currentGame.gameSettings.length);

    final List<String> tips = List<String>.filled(wordLength, '', growable: false);

    if ((!state.currentGame.checkWordIsValid(candidate))) {
      return List<String>.filled(wordLength, 'wrong', growable: false);
    }

    String replaceCharAt(String oldString, int index, String newChar) {
      return oldString.substring(0, index) + newChar + oldString.substring(index + 1);
    }

    // Check correctly placed letters
    for (int i = 0; i < wordLength; i++) {
      if ((tips[i] == '') && (word[i] == candidate[i])) {
        addFoundLetter(word[i], i);
        word = replaceCharAt(word, i, ' ');
        candidate = replaceCharAt(candidate, i, ' ');
        tips[i] = 'good';
      }
    }

    // Check misplaced letters
    for (int i = 0; i < wordLength; i++) {
      for (int j = 0; j < wordLength; j++) {
        if ((candidate[j] != ' ') && (candidate[j] == word[i])) {
          word = replaceCharAt(word, i, ' ');
          candidate = replaceCharAt(candidate, j, ' ');
          tips[j] = 'misplaced';
        }
      }
    }

    return tips;
  }

  @override
  GameState? fromJson(Map<String, dynamic> json) {
    final Game currentGame = json['currentGame'] as Game;

    return GameState(
      currentGame: currentGame,
    );
  }

  @override
  Map<String, dynamic>? toJson(GameState state) {
    return <String, dynamic>{
      'currentGame': state.currentGame.toJson(),
    };
  }
}
